package memtrace_viewer

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Bonsai_simple_table.MakeSource

Parameters

module Row : Row
module Col_id : Id

Signature

Sourcetype cell = {
  1. node : Bonsai_web.Vdom.Node.t;
  2. attrs : Bonsai_web.Vdom.Attr.t list;
}

This will be turned into a td with contents node and attributes attrs.

Sourcemodule Column : sig ... end
Sourcemodule Model : sig ... end
Sourcemodule Input : sig ... end

In addition to specifying the full range of rows and cols, we also ask for the IDs in order. This can be used to implement both filtering and sorting (although there is no support within this component to e.g. sort on a column when the user clicks on the header).

Sourcemodule Action : sig ... end
Sourcemodule Result : sig ... end
Sourceval bonsai : Input.t Bonsai_web.Value.t -> Result.t Bonsai_web.Computation.t